#' Reconcile locol interaction networks
#'
#' Function which takes the output of \code{\link{generate_JRIM_models}} and
#' assembles the connections into a data frame with JRIM co-accessibility
#' scores. Modified from function \code{cicero::assemble_connections}.
#'
#'  
#' @param JRIM_model_list A list of JRIM output objects, generally, the
#'   output of \code{\link{generate_JRIM_models}}.
#' @param K The number of tissues.
#' @param verbose Logical, should the function run silently?
#'
#' @return A list of data frame of connections with their cicero co-accessibility
#'   scores of all K tissues.
#' @export
#'
assemble_connections_JRIM <- function(JRIM_model_list, K, verbose = TRUE) 
{
  types <- sapply(JRIM_model_list, class)
  char_hbn <- JRIM_model_list[types=="character"]
  gl_only <- JRIM_model_list[types=="jgl"]
  if(verbose) {
    print(paste("Successful JGL cicero models: ", length(gl_only)))
    print("Other models: ")
    print(table(unlist(char_hbn)))
    print(paste("Models with errors: ", sum(is.null(JRIM_model_list))))
  }
  cors <- list()
  for (i in 1:K) {
    cors[[i]] <- lapply(gl_only, function(gl) {
      cors <- stats::cov2cor(gl$theta[[i]])
      data.table::melt(cors)
      })
  }
  reconcile_values <- function(values){ 
    # Borrowed from Cicero package.
    if (length(values) == 1) return(values)
    if (sum(values >= 0) == length(values)) return(mean(values))
    if (sum(values <= 0) == length(values)) return(mean(values))
    if (sum(values == 0) == length(values)) return(0)
    return(NA_real_)
  }
  cors_rec <- list()
  for (i in 1:K) { 
    cors[[i]] <- data.table::rbindlist(cors[[i]]) 
    data.table::setkey(cors[[i]], "Var1", "Var2")
	temp_cors = tidyr::unite(cors[[i]], "Var_conn", Var1, Var2,sep=',')
	temp_groupby <- dplyr::group_by(temp_cors, Var_conn)
	temp_rec <- dplyr::summarise(temp_groupby, count = reconcile_values(value))
	temp_rec= tidyr::separate(temp_rec, Var_conn, c('Var1', 'Var2'), sep=',')
    #cors_rec[[i]] <- as.data.frame(temp_cors[,list(mean_coaccess = reconcile_values(value)), by="Var1,Var2"])
	cors_rec[[i]] = as.data.frame(temp_rec)
    names(cors_rec[[i]]) <- c("Peak1", "Peak2", "coaccess")
    cors_rec[[i]] <- cors_rec[[i]][cors_rec[[i]]$Peak1 != cors_rec[[i]]$Peak2,]
    cors_rec[[i]] <- cors_rec[[i]][cors_rec[[i]]$coaccess!=0,]
  }
  ##multi (-1) to get partial correlation
  for (i in 1:K) {
    cors_rec[[i]]$coaccess = (-1) * cors_rec[[i]]$coaccess
  }
  return(cors_rec)
}