#' Calculate scaling parameters
#'
#' Function to calculate scaling parameters lambda1 and lambda2
#' for random genomic windows. 
#'
#' @param cds_list A list of JRIM CDS object generated using \code{\link{make_JRIM_cds}}.
#' @param s Power law value. See details for more information.
#' @param maxit Maximum number of iterations for distance_parameter estimation.
#' @param weights 'equal' or 'sample.size'. The selection of n_k in joint graphical lasso model.
#' when weights='equal', n_k=1 for all (k in 1:K). when weights='sample.size', n_k=the group 
#' number of tissue k. 
#' @param w2_initial The initial value of reparameterized parameter w_2.
#' @param window Size of the genomic window to query, in base pairs.
#' @param sample_num Number of random windows to calculate
#'   \code{distance_parameter} for.
#' @param max_elements Maximum number of elements per window allowed. Prevents
#'   very large models from slowing performance.
#' @param penalize.diagonal Logical, should the diagonal elements be penalized? See JGL packages for details.
#' @param distance_constraint Maximum distance of expected connections. Must be
#'   smaller than \code{window}.
#' @param w1_convergence Convergence step size for w_1 calculation.
#' @param w2_convergence Convergence step size for w_2 calculation.
#' @param verbose Logical, should the function run silently?
#' @param genomic_coords Either a data frame or a path (character) to a file
#'   with chromosome lengths. The file should have two columns, the first is
#'   the chromosome name (ex. "chr1") and the second is the chromosome length
#'   in base pairs. See \code{data(human.hg19.genome)} for an example. If a
#'   file, should be tab-separated and without header.
#' @param sparsity_constrain The requirement of sparsity in each local window.
#' @param long_dist_constrain The requirement of sparsity of long distance interactions in each local window.
#' @param similarity_constrain The requirement of similarity across networks in each local window.
#'
#' @return A list of results of length \code{sample_num} and the resulting value of lambda1 lambda2.
#'
#' @export
#'
estimate_JRIM_parameters <- function(cds_list, 
                                     s=0.75,
                                     maxit=50,
                                     weights="equal",
                                     w2_initial=0.1,
                                     window=500000,
                                     sample_num=50,
                                     max_elements=200,
                                     penalize.diagonal=FALSE,
                                     distance_constraint=250000,
                                     w1_convergence=1e-4,
                                     w2_convergence=1e-4,
                                     verbose=TRUE,
                                     genomic_coords=cicero::human.hg19.genome,
                                     sparsity_constrain=0.1,
                                     long_dist_constrain=0.05,
                                     similarity_constrain=0.6) 
{
  K <- length(cds_list)
  grs <- generate_windows(window, genomic_coords)
  for (i in 1:K){
    fData(cds_list[[i]])$chr <- gsub("chr", "", fData(cds_list[[i]])$chr)
    fData(cds_list[[i]])$bp1 <- as.numeric(as.character(fData(cds_list[[i]])$bp1))
    fData(cds_list[[i]])$bp2 <- as.numeric(as.character(fData(cds_list[[i]])$bp2))
  }


  it <- 0
  w1_parameters <- list()
  w1_parameters_calced <- 0
  if (verbose){
    message("Finding w1...")
  }


  while(sample_num > w1_parameters_calced & it < 3 * sample_num) {
    it <- it + 1
    win <- sample(1:length(grs),1)
    win_range_list <- list()
    peaks_list <- list()
    for (i in 1:K) {
      win_range_list[[i]] <- get_genomic_range(grs, cds_list[[i]], win)
      peaks_list[[i]] <- fData(win_range_list[[i]])$site_name 
    }
    common_peaks <- Reduce(intersect, peaks_list)
    if (length(common_peaks)<=1) {
      next()
    }
    if (length(common_peaks) > max_elements) {
      next()
    }
    dist_matrix <- calc_dist_matrix(win_range_list[[1]])
    dist_matrix <- dist_matrix[common_peaks, common_peaks]
    exprs_list <- list()
    for (i in 1:K) { 
      exprs_list[[i]] <- t(exprs(win_range_list[[i]])[common_peaks,]) 
    }
    

    #if (penalty=="group"){
      w1_ <- find_JRIM_parameter(exprs_list,dist_matrix,w_fix=w2_initial,weights=weights,
                                 maxit = maxit, s=s, distance_constraint = distance_constraint,
                                 fixed="w2",penalize.diagonal=penalize.diagonal,verbose=FALSE,
                                 parameter_convergence = w1_convergence,
                                 sparsity_constrain=sparsity_constrain,
                                 long_dist_constrain=long_dist_constrain,
                                 similarity_constrain=similarity_constrain)

    #}

    
    if (class(w1_) != "numeric") next()
    if (verbose){
      message("------win:", win, ", -----", length(common_peaks),  "------" , "w1:",w1_)
    }

    w1_parameters = c(w1_parameters, w1_)
    w1_parameters_calced <- w1_parameters_calced + 1
  }



  w1_parameters <- unlist(w1_parameters)
  mean_w1 <- mean(w1_parameters)

  if (verbose){
    message("---Estimated w1:", mean_w1)
  }



  ###Estimating w2
  it <- 0
  w2_parameters <- list()
  w2_parameters_calced <- 0
  if (verbose){
    message("Finding w2...")
  }
  while(sample_num > w2_parameters_calced & it < 3 * sample_num) {
    it <- it + 1
    win <- sample(1:length(grs),1)
    win_range_list <- list()
    peaks_list <- list()
    for (i in 1:K) {
      win_range_list[[i]] <- get_genomic_range(grs, cds_list[[i]], win)
      peaks_list[[i]] <- fData(win_range_list[[i]])$site_name 
    }
    common_peaks <- Reduce(intersect, peaks_list)
    if (length(common_peaks)<=1) {
      next()
    }
    if (length(common_peaks) > max_elements) {
      next()
    }
    dist_matrix <- calc_dist_matrix(win_range_list[[1]])
    dist_matrix <- dist_matrix[common_peaks, common_peaks]
    exprs_list <- list()
    for (i in 1:K) { 
      exprs_list[[i]] <- t(exprs(win_range_list[[i]])[common_peaks,]) 
    }

    #if (penalty=="group"){
      w2_ <- find_JRIM_parameter(exprs_list,dist_matrix,w_fix=mean_w1,weights=weights,
                                 maxit = maxit, s=s, distance_constraint = distance_constraint,
                                 fixed="w1",penalize.diagonal=penalize.diagonal,verbose=FALSE,
                                 parameter_convergence = w2_convergence,
                                 sparsity_constrain=sparsity_constrain,
                                 long_dist_constrain=long_dist_constrain,
                                 similarity_constrain=similarity_constrain)
    #}


    if (class(w2_) != "numeric") {next()}
    if (verbose){
      message("------win:", win, ", -----", length(common_peaks),  "------" , "w2:",w2_)
    }

    w2_parameters = c(w2_parameters, w2_)
    w2_parameters_calced <- w2_parameters_calced + 1
  }

  w2_parameters <- unlist(w2_parameters)
  mean_w2 <- mean(w2_parameters)

  if (verbose){
    message("---Estimated w2:", mean_w2)
  }

  lambda1_calced = mean_w1 * (1 - mean_w2)
  lambda2_calced = sqrt(2) * mean_w1 * mean_w2
  parameters <- list(lambda1=lambda1_calced, lambda2=lambda2_calced, w1=mean_w1, w2=mean_w2,
                     w1_list=w1_parameters, w2_list=w2_parameters)
  return(parameters)
}
