#' Generate JRIM models in local genomic windows
#'
#' Function to adopt joint graphical lasso models on 
#' overlapping genomic windows.
#'
#' @param cds_list JRIM CDS objects of K tissues generated using \code{\link{make_JRIM_cds}}.
#' @param lambda1 Scaling parameter value of l1 norm penalty. Generally,
#'   the mean of the calculated lambda1 values from
#'   \code{\link{estimate_JRIM_parameters}}.
#' @param lambda2 Scaling parameter value of group lasso penalty. Generally,
#'   the mean of the calculated lambda2 values from
#'   \code{\link{estimate_JRIM_parameters}}.
#' @param weights 'equal' or 'sample.size'. The selection of n_k in joint graphical lasso model.
#' when weights='equal', n_k=1 for all (k in 1:K). when weights='sample.size', n_k=the group 
#' number of tissue k. 
#' @param s Power law value of distance penalty.
#' @param window Size of the genomic window to query, in base pairs.
#' @param maxiter Maximum iteration number of joint graphical lasso model. 
#' See JGL packages for details.
#' @param max_elements Maximum number of elements per window allowed. Prevents
#'   very large models from slowing performance.
#' @param tol The convergence parameter of joint graphical lasso model. See JGL packages for details.
#' @param verbose Logical, should the function run silently?
#' @param penalize.diagonal Logical, should the diagonal elements be penalized? See JGL packages for details.
#' @param genomic_coords Either a data frame or a path (character) to a file
#'   with chromosome lengths. The file should have two columns, the first is
#'   the chromosome name (ex. "chr1") and the second is the chromosome length
#'   in base pairs. See \code{data(human.hg19.genome)} for an example. If a
#'   file, should be tab-separated and without header.
#'
#'
#' @return A list of results for each window. This list can be
#'   directly input into \code{\link{assemble_connections_JRIM}} 
#'   to create regulatory interaction maps of different cell populatins.
#' @export
#'
generate_JRIM_models <- function(cds_list,
                                 lambda1,
                                 lambda2,
                                 weights="equal",
                                 s = 0.75,
                                 window = 500000,
                                 maxiter = 500,
                                 max_elements = 300,
                                 tol=1e-5,
                                 verbose = FALSE,
                                 penalize.diagonal=FALSE,
                                 genomic_coords = cicero::human.hg19.genome) 
{
  assertthat::assert_that(weights %in% c("equal", "sample.size"))
  assertthat::assert_that(assertthat::is.number(lambda1))
  assertthat::assert_that(assertthat::is.number(lambda2))
  assertthat::assert_that(assertthat::is.number(s), s < 1, s > 0)
  assertthat::assert_that(assertthat::is.number(window))
  assertthat::assert_that(assertthat::is.number(maxiter))
  assertthat::assert_that(assertthat::is.count(max_elements))

  K <- length(cds_list)
  grs <- generate_windows(window, genomic_coords)
  for (i in 1:K) {
    fData(cds_list[[i]])$chr <- gsub("chr", "", fData(cds_list[[i]])$chr)
    fData(cds_list[[i]])$bp1 <- as.numeric(as.character(fData(cds_list[[i]])$bp1))
    fData(cds_list[[i]])$bp2 <- as.numeric(as.character(fData(cds_list[[i]])$bp2))
  }
  

  outlist <- list()
  for (win in 1:length(grs)){
    if (verbose) {
      if (win%%50==0) {message("Dealing with NO.",win, " window.")}
    }
    jgl.results <- "Error"
    win_range_list <- list()
    peaks_list <- list()
    for (i in 1:K){
      win_range_list[[i]] <- get_genomic_range(grs, cds_list[[i]], win)
      peaks_list[[i]] <- fData(win_range_list[[i]])$site_name 
    }
    common_peaks <- Reduce(intersect, peaks_list)

    if (length(common_peaks)<=1) {
      outlist[[win]] <- "Zero or one element in range"
      next()
    }
    if (length(common_peaks) > max_elements) {
      outlist[[win]] <- "Too many elements in range"
      next()
    }
    dist_matrix <- calc_dist_matrix(win_range_list[[1]])
    dist_matrix <- dist_matrix[common_peaks, common_peaks]
    rho_mat <- get_rho_mat(dist_matrix, lambda1, s)
    exprs_list <- list()
    for (i in 1:K){
      exprs_list[[i]] <- t(exprs(win_range_list[[i]])[common_peaks,])
    }
    jgl.results = JGL_for_ATAC(Y=exprs_list,penalty="group",lambda1=rho_mat,lambda2=lambda2,weights=weights,
                               penalize.diagonal=penalize.diagonal,tol=tol,maxiter=maxiter,
                               return.whole.theta=TRUE)
                                  
    ####message error
    if (is.character(jgl.results)){
      if (verbose){
        message(win,"-----error:",jgl.results)      
      }
    }
    outlist[[win]] <- jgl.results
  }

  names_df <- as.data.frame(grs)
  names(outlist) <- paste(names_df$seqnames,
                          names_df$start,
                          names_df$end, sep="_")
  return(outlist)
}