#' Create JRIM input CDS
#'
#' Function to generate aggregated input CDSs for JRIM. \code{generate_JRIM_models}
#' takes as input aggregated CDSs object. This function will group similar cells
#' per tissue using a KNN method and aggregate their binary accessibility scores 
#' if they do not overlap any existing group with more than 80% cells. 
#' Modified from function \code{cicero::make_cicero_cds}.
#'
#' @param cds List of input CDS objects of K tissues.
#' @param reduced_coordinates A data frame with columns representing the
#'   coordinates of each cell in reduced dimension space (generally 2-3
#'   dimensions).
#' @param k_neigh Number of cells to aggregate per group.
#' @param size_factor_normalize Logical, should accessibility values be
#'   normalized by size factor?
#' @param verbose Logical, should warning and info messages be printed?
#' @param max_overlap The maximum overlapping ratio of two groups.
#'
#'
#' @return Aggregated CDS objects of all K tissues.
#' @export
#'
make_JRIM_cds <- function (cds, reduced_coordinates, k_neigh = 50, 
          size_factor_normalize = TRUE, verbose = TRUE, max_overlap=0.8) 
{ #modified from JGL package.
  assertthat::assert_that(class(cds) == "CellDataSet")
  assertthat::assert_that(is.data.frame(reduced_coordinates) | 
                            is.matrix(reduced_coordinates))
  assertthat::assert_that(assertthat::are_equal(nrow(reduced_coordinates), 
                                                nrow(pData(cds))))
  assertthat::assert_that(setequal(row.names(reduced_coordinates), 
                                   colnames(cds)))
  assertthat::assert_that(assertthat::is.count(k_neigh) & k_neigh > 1)

  assertthat::assert_that(is.logical(size_factor_normalize))
  assertthat::assert_that(is.logical(verbose))

  reduced_coordinates <- as.data.frame(reduced_coordinates)
  reduced_coordinates <- reduced_coordinates[colnames(cds), ]

  # Create a k-nearest neighbors map
  nn_map <- as.data.frame(FNN::knn.index(reduced_coordinates, 
                                         k = (k_neigh - 1)))

  row.names(nn_map) <- row.names(reduced_coordinates)
  nn_map$agg_cell <- 1:nrow(nn_map)
  good_choices <- 1:nrow(nn_map)

  if (verbose)  
    message("Sample cells randomly.")

  #Sample cells randomly
  choice <- sample(1:length(good_choices), size = 1, replace = FALSE)
  chosen <- good_choices[choice]
  good_choices <- good_choices[good_choices != good_choices[choice]]

  it <- 0
  ##Slow (contain calculating of overlapping between cell groups)
  while (length(good_choices) > 0 & it < 5000) {
    it <- it + 1
    choice <- sample(1:length(good_choices), size = 1, replace = FALSE)
    new_chosen <- c(chosen, good_choices[choice])
    good_choices <- good_choices[good_choices != good_choices[choice]]
    cell_sample <- nn_map[new_chosen, ]

    ##calculate overlapping between cell groups
    combs <- data.frame(1:(nrow(cell_sample) - 1), nrow(cell_sample))
    shared <- apply(combs, 1, function(x) {    #Slow
      (k_neigh * 2) - length(unique(as.vector(as.matrix(cell_sample[x, 
                                                              ]))))
    })
    if (max(shared) < max_overlap * k_neigh) {
      chosen <- new_chosen
    }
  }


  if (verbose) {
    cell_sample <- nn_map[chosen, ]
    combs <- t(combn(nrow(cell_sample), 2))
    shared <- apply(combs, 1, function(x) {
     (k_neigh * 2) - length(unique(as.vector(as.matrix(cell_sample[x, 
                                                              ]))))
    })
    message(paste0("Overlap QC metrics:\nCells per bin: ", 
                   k_neigh, "\nMaximum shared cells bin-bin: ", max(shared), 
                   "\nMean shared cells bin-bin: ", mean(shared), "\nMedian shared cells bin-bin: ", 
                   median(shared)))
    #if (mean(shared)/k_neigh > 0.1) 
    #  warning("On average, more than 10% of cells are\n                                   shared between paired bins.")
  }

  #aggregating binary counts of cells within one group
  exprs_old <- exprs(cds)
  
  mask <- sapply(seq_len(nrow(cell_sample)), function(x) seq_len(ncol(exprs_old)) %in% cell_sample[x,,drop=FALSE])
  mask <- Matrix::Matrix(mask)
  new_exprs <- exprs_old %*% mask
  
  new_exprs <- Matrix::t(new_exprs)
  new_exprs <- as.matrix(new_exprs)

  pdata <- pData(cds)
  new_pdata <- plyr::adply(cell_sample, 1, function(x) {
    sub <- pdata[as.numeric(x[1, ]), ]
    df_l <- list()
    df_l["temp"] <- 1
    data.frame(df_l)
  })
  new_pdata <- new_pdata[, (k_neigh):ncol(new_pdata)]
  new_pdata$agg_cell <- paste("agg", new_pdata$agg_cell, sep = "")
  row.names(new_pdata) <- new_pdata$agg_cell
  row.names(new_exprs) <- new_pdata$agg_cell
  new_exprs <- as.matrix(t(new_exprs))
  fdf <- fData(cds)
  new_pdata$temp <- NULL
  fd <- new("AnnotatedDataFrame", data = fdf)
  pd <- new("AnnotatedDataFrame", data = new_pdata)
  JRIM_cds <- suppressWarnings(newCellDataSet(new_exprs, 
                                                phenoData = pd, featureData = fd, expressionFamily = negbinomial.size(), 
                                                lowerDetectionLimit = 0))
  JRIM_cds <- monocle::detectGenes(JRIM_cds, min_expr = 0.1)
  JRIM_cds <- BiocGenerics::estimateSizeFactors(JRIM_cds)
  JRIM_cds <- suppressWarnings(BiocGenerics::estimateDispersions(JRIM_cds))
  if (any(!c("chr", "bp1", "bp2") %in% names(fData(JRIM_cds)))) {
    fData(JRIM_cds)$chr <- NULL
    fData(JRIM_cds)$bp1 <- NULL
    fData(JRIM_cds)$bp2 <- NULL
    fData(JRIM_cds) <- cbind(fData(JRIM_cds), df_for_coords(row.names(fData(JRIM_cds))))
  }
  if (size_factor_normalize) {
    Biobase::exprs(JRIM_cds) <- t(t(Biobase::exprs(JRIM_cds))/Biobase::pData(JRIM_cds)$Size_Factor)
  }
  JRIM_cds
}