#' Plot connections of multiple tissues
#'
#' Plotting function for JRIM connections of multiple tissues. Uses \code{\link[Gviz]{plotTracks}}
#' as its basis
#'
#' @param connections_list Data frame of connections of K tissues, which must include the
#'   columns 'Peak1', 'Peak2', and 'coaccess'. 
#' @param K The number of tissues.
#' @param chr The chromosome of the region you would like to plot in the form
#'   'chr10'.
#' @param minbp The base pair coordinate of the start of the region to be
#'   plotted.
#' @param maxbp The base pair coordinate of the end of the region to be plotted.
#' @param fontsize The size of fontsize.
#' @param coaccess_cutoff The minimum co-accessibility score you would
#'   like to be plotted. Default is 0. It can be a number or a list of number in length K.
#' @param peak_color Color for peak annotations - a single color, the name of a
#'   column containing color values that correspond to Peak1, or the name of
#'   column containing a character or factor to base peak colors on.
#' @param connection_color Color for connection lines. A single color, the name
#'   of a column containing color values, or the name of a column containing a
#'   character or factor to base connection colors on.
#' @param alpha_by_coaccess Logical, should the transparency of connection
#'   lines be scaled based on co-accessibility score?
#' @param connection_width Width of connection lines.
#' @param connection_ymax Connection y-axis height. If \code{NULL}, chosen
#'   automatically. It can be a number or a list of number in length K.
#' @param gene_model Either \code{NULL} or a data.frame. The data.frame should
#'   be in a form compatible with the Gviz function
#'   \code{\link[Gviz]{GeneRegionTrack-class}} (cannot have NA as column names).
#' @param gene_model_size The size of plotted gene model.
#' @param gene_model_color Color for gene annotations.
#' @param gene_model_shape Shape for gene models, passed to
#'   \code{\link[Gviz]{GeneRegionTrack-class}}. Options described at
#'   \code{\link[Gviz]{GeneRegionTrack-class}}.
#' @param collapseTranscripts Logical or character scalar. Can be one in
#'   \code{gene}, \code{longest}, \code{shortest} or \code{meta}. Variable is
#'   passed to the \code{\link[Gviz]{GeneRegionTrack-class}} function of Gviz.
#'   Determines whether and how to collapse related transcripts. See Gviz
#'   documentation for details.
#' @param include_axis_track Logical, should a genomic axis be plotted?
#' @param return_as_list Logical, if TRUE, the function will not plot, but will
#'   return the plot components as a list. Allows user to add/customize Gviz
#'   components and plot them separately using \code{\link[Gviz]{plotTracks}}.
#' @param viewpoint \code{NULL} or Coordinates in form "chr1_10000_10020". Use
#'   viewpoint if you would like to plot cicero connections "4C-seq style".
#'   Only connections originating in the viewpoint will be shown. Ideal for
#'   comparisons with 4C-seq data. If comparison_viewpoint is \code{TRUE}, any
#'   comparison track will be subsetted as well.
#' @param comparison_viewpoint Logical, should viewpoint apply to comparison
#'   track as well?
#' @param viewpoint_color Color for the highlight border.
#' @param viewpoint_fill Color for the highlight fill.
#' @param viewpoint_alpha Alpha value for the highlight fill.
#'
#' @return A gene region plot, or list of components if return_as_list is
#'   \code{TRUE}.
#' @export
#' @import Gviz
#'
plot_multiple_connections <- function(connections_list,
                                      K,
                                      chr,
                                      minbp,
                                      maxbp,
                                      fontsize = 6,
                                      coaccess_cutoff = 0,
                                      peak_color = "#B4656F",
                                      connection_color = "#7F7CAF",
                                      connection_color_legend = TRUE,
                                      alpha_by_coaccess = FALSE,
                                      connection_width = 2,
                                      connection_ymax = NULL,
                                      gene_model = NULL,
                                      gene_model_size = 1,
                                      gene_model_color = "#81D2C7",
                                      gene_model_shape = c("smallArrow", "box"),
                                      collapseTranscripts = FALSE,
                                      include_axis_track = TRUE,
                                      axis_track_size = 0.2,
                                      return_as_list = FALSE,
                                      viewpoint = NULL,
                                      viewpoint_color = "#F0544F",
                                      viewpoint_fill = "#EFD8D7",
                                      viewpoint_alpha = 0.5) 
{ # modified from JGL package.
  # Check inputs:
  for (i in 1:K){
    assertthat::assert_that(is.data.frame(connections_list[[i]]))
    assertthat::assert_that(assertthat::has_name(connections_list[[i]], "Peak1"),
                          assertthat::has_name(connections_list[[i]], "Peak2"),
                          assertthat::has_name(connections_list[[i]], "coaccess"))
  }
  assertthat::assert_that(assertthat::is.number(K))
  assertthat::assert_that(K==length(all_cons))

  assertthat::assert_that(is.character(chr), is_chr(chr))
  assertthat::assert_that(assertthat::is.number(minbp),
                          assertthat::is.number(maxbp))

  if (length(coaccess_cutoff) == 1) {
    coaccess_cutoff = vector(mode = "numeric", K) + coaccess_cutoff
  }
  assertthat::assert_that(K==length(coaccess_cutoff))

  for (i in 1:K){
    assertthat::assert_that(assertthat::is.number(coaccess_cutoff[i]))
    assertthat::assert_that(coaccess_cutoff[i] >=0)
  }



  if (length(peak_color)==1) {
    assertthat::assert_that(is_color(peak_color))
    color_list = c()
    for (i in 1:K){
      color_list[i] = peak_color
    }
    peak_color = color_list
  }
  assertthat::assert_that(length(peak_color) == K)

  if (length(connection_color)==1) {
    assertthat::assert_that(is_color(connection_color))
    color_list = c()
    for (i in 1:K){
      color_list[i] = connection_color
    }
    connection_color = color_list
  }
  assertthat::assert_that(length(connection_color) == K)



  assertthat::assert_that(is.logical(alpha_by_coaccess))

  assertthat::assert_that(is.numeric(connection_width), connection_width > 0)


  if (!is.null(connection_ymax)){
    if (length(connection_ymax) == 1){
      connection_ymax = vector(mode = "numeric", K) + connection_ymax
    }
    assertthat::assert_that(K==length(connection_ymax))
    for (i in 1:K){
      assertthat::assert_that(assertthat::is.number(connection_ymax[i]))
      assertthat::assert_that(connection_ymax[i] > 0)
    }
  }

  if (!is.null(gene_model)) {
    assertthat::assert_that(is.data.frame(gene_model),
                            assertthat::has_name(gene_model, "chromosome"),
                            assertthat::has_name(gene_model, "start"),
                            assertthat::has_name(gene_model, "end"),
                            assertthat::has_name(gene_model, "strand"),
                            assertthat::has_name(gene_model, "transcript"))
    assertthat::assert_that(is_color(gene_model_color))
  }


  assertthat::assert_that(is.logical(include_axis_track),
                          is.logical(return_as_list))

  assertthat::assert_that(is.logical(collapseTranscripts) |
                            is.character(collapseTranscripts))
  if(is.character(collapseTranscripts)) {
    assertthat::assert_that(collapseTranscripts %in% c("gene", "longest",
                                                       "shortest", "meta"))
  }

  if (!is.null(viewpoint)) {
    assertthat::assert_that(is.character(viewpoint))
    assertthat::assert_that(is_color(viewpoint_color))
    assertthat::assert_that(is_color(viewpoint_fill))
    assertthat::assert_that(assertthat::is.number(viewpoint_alpha),
                            viewpoint_alpha > 0,
                            viewpoint_alpha <= 1)
    viewpoint <- gsub(",", "", viewpoint)
    viewpoint <- gsub(":|-", "_", viewpoint)
  }


  sub = list()
  anntrack = list()
  connection_ymax_cal = c()
  for (i in 1:K){
    # assign peak colors
    connections_list[[i]]$peak_color <- peak_color[[i]]

    sub[[i]] <- generate_plotting_subset(connections_list[[i]], chr, minbp, maxbp)

    if (nrow(sub[[i]]) == 0) stop("No peaks in the specified range, nothing to plot!")
    sub[[i]]$peak_color <- peak_color[[i]]

    anntrack[[i]] <- make_peak_track(sub[[i]])
    sub[[i]] <- sub[[i]][sub[[i]]$coaccess >= coaccess_cutoff[[i]] &
                      sub[[i]]$coaccess > 0 &
                      !is.na(sub[[i]]$coaccess),]

    if (!is.null(viewpoint)) {
      viewpoint_overs <- find_overlapping_coordinates(c(sub[[i]]$Peak1, sub[[i]]$Peak2),
                                                      viewpoint)
      sub[[i]] <- sub[[i]][sub[[i]]$Peak1 %in% viewpoint_overs | sub[[i]]$Peak2 %in% viewpoint_overs,]
      if (nrow(sub[[i]]) == 0) warning("No connections in viewpoint in tissue ",i)
    }


    if (!nrow(sub[[i]]) == 0) {
      sub[[i]]$color <- connection_color[[i]]

      sub[[i]]$width <- connection_width

      sub[[i]] <- sub[[i]][,c("chr", "bp1", "bp2", "chr_2", "bp1_2", "bp2_2", "coaccess",
                    "width", "color")]
      names(sub[[i]]) <- c("chrom1","start1","stop1","chrom2","start2","stop2",
                      "height", "width", "color")
    } else {
      warning("No connections above coaccess_cutoff in tissue ",i)
    }
    if (is.null(connection_ymax)) {
      connection_ymax_cal[i] <- signif_up(max(sub[[i]]$height, na.rm=TRUE))
      if (is.na(connection_ymax_cal[i]) | connection_ymax_cal[i] <= 0) {
        connection_ymax_cal[i] <- 1
        warning("connection_ymax calc failed in tissue ", i)
      }
    }
  }
  if (is.null(connection_ymax)) {connection_ymax = connection_ymax_cal}
  assertthat::assert_that(K==length(connection_ymax))


  size_track = c()
  component_num = 0
  component_list = list()
  ctrack = list()
  for (i in 1:K){
    ctrack[[i]] = CustomTrack(plottingFunction=function(GdObject, prepare) {
      Gviz::displayPars(GdObject) <- list(ylim = c(0,connection_ymax[plot_it]))
      if(!prepare) {
        plotBedpe(sub[[plot_it]], chrom = chr, chromstart=minbp, chromend=maxbp,
                  connection_ymax[plot_it], coaccess_cutoff[plot_it],
                  connection_width, alpha_by_coaccess)
        #Global variable
        plot_it <<- plot_it + 1
       }
      return(invisible(GdObject))}, name="", fontsize.group=fontsize,fontsize=fontsize)

    component_num = component_num + 1
    component_list[[component_num]] = ctrack[[i]]
    size_track = c(size_track, 1)

    component_num = component_num + 1
    component_list[[component_num]] = anntrack[[i]]
    size_track = c(size_track, .2)
  }
  plot_it <<- 1




if (include_axis_track) {
    atrack <- Gviz::GenomeAxisTrack(fontsize=fontsize)
    size_track <- c(size_track, axis_track_size)
    component_num <- component_num + 1
    component_list[[component_num]] <- atrack
  }



  if(!is.null(gene_model)) {
    gene_model <-
      gene_model[gene_model$chromosome == chr &
                   ((gene_model$start > minbp & gene_model$start < maxbp) |
                    (gene_model$end > minbp & gene_model$end < maxbp) |
                    (gene_model$start < minbp & gene_model$end > maxbp)),]
    # Define gene model track
    grtrack <- make_gene_model_track(gene_model, chr,
                                     collapseTranscripts, gene_model_color,
                                     gene_model_shape)
    component_num <- component_num + 1
    component_list[[component_num]] <- grtrack
    size_track <- c(size_track, gene_model_size)
  }

  if(!is.null(viewpoint)) {
    view_chr <- stringr::str_split_fixed(viewpoint, "_", 3)[,1]
    if (chr != view_chr) warning("Viewpoint not on correct chromosome")
    else {
      view_start <- stringr::str_split_fixed(viewpoint, "_", 3)[,2]
      view_end <- stringr::str_split_fixed(viewpoint, "_", 3)[,3]
      ht1 <- Gviz::HighlightTrack(trackList = component_list,
                                  start = as.numeric(view_start),
                                  end = as.numeric(view_end),
                                  chromosome = chr,
                                  col = viewpoint_color,
                                  fill = viewpoint_fill,
                                  inBackground = FALSE,
                                  alpha = viewpoint_alpha)
      component_list <- ht1
    }
  }

  if (return_as_list) return(component_list)
  return(plotTracks(component_list, title.width = .5, showTitle = TRUE,
              from = minbp, to = maxbp, chromosome = chr, sizes = size_track,
             transcriptAnnotation = "symbol", background.title = "transparent",
             col.border.title="transparent", lwd.border.title = "transparent",
              col.axis = "black", fontsize.group = 6,#col.title="white",
             fontcolor.legend = "black"))
}
